@@ -112,19 +112,8 @@ class ItemEditor(admin.ModelAdmin):
         if res:
             return self._frontend_editing_view(request, res.group(1), res.group(2), res.group(3))
 
-        ModelForm = modelform_factory(self.model, exclude=('parent',),
-            formfield_callback=self._formfield_callback(request=request),
-            form=self.form)
-
-        # generate a formset type for every concrete content type
-        inline_formset_types = [(
-            content_type,
-            inlineformset_factory(self.model, content_type, extra=1,
-                form=getattr(content_type, 'feincms_item_editor_form', ItemEditorForm),
-                formfield_callback=self._formfield_callback(request=request))
-            ) for content_type in self.model._feincms_content_types]
-
         opts = self.model._meta
+
         try:
             obj = self.model._default_manager.get(pk=unquote(object_id))
         except self.model.DoesNotExist:
@@ -133,7 +122,43 @@ class ItemEditor(admin.ModelAdmin):
         if not self.has_change_permission(request, obj):
             raise PermissionDenied
 
+        messages = list()
+
+        if "revision" in request.GET:
+            from reversion.models import Revision
+
+            try:
+                revision = Revision.objects.get(pk=request.GET['revision'])
+            except Revision.DoesNotExist:
+                raise Http404
+
+            messages.append(_('Click save to replace the current content with this version'))
+        else:
+            revision = None
+
+        ModelForm = self.get_form(
+            request,
+            obj,
+            # NOTE: Fields *MUST* be set to None to avoid breaking
+            # django.contrib.admin.option.ModelAdmin's default get_form()
+            # will generate a very abbreviated field list which will cause
+            # KeyErrors during clean() / save()
+            fields=None,
+            formfield_callback=self._formfield_callback(request=request),
+            form=self.form
+        )
+
+        # generate a formset type for every concrete content type
+        inline_formset_types = [(
+            content_type,
+            inlineformset_factory(self.model, content_type, extra=1,
+                fk_name='parent', #added so multiple foreign keys are not a problem
+                form=getattr(content_type, 'feincms_item_editor_form', ItemEditorForm),
+                formfield_callback=self._formfield_callback(request=request))
+            ) for content_type in self.model._feincms_content_types]
+
         if request.method == 'POST':
+            FORM_DATA = {}
             model_form = ModelForm(request.POST, request.FILES, instance=obj)
 
             inline_formsets = [
@@ -142,9 +167,10 @@ class ItemEditor(admin.ModelAdmin):
                 for content_type, formset_class in inline_formset_types]
 
             if model_form.is_valid() and all_valid(inline_formsets):
-                model_form.save()
+                model_form.save(commit=False)
                 for formset in inline_formsets:
                     formset.save()
+                model_form.save(commit=True)
 
                 msg = _('The %(name)s "%(obj)s" was changed successfully.') % {'name': force_unicode(opts.verbose_name), 'obj': force_unicode(obj)}
                 if request.POST.has_key("_continue"):
@@ -156,11 +182,55 @@ class ItemEditor(admin.ModelAdmin):
                 else:
                     self.message_user(request, msg)
                     return HttpResponseRedirect("../")
+        elif revision:
+            FORM_DATA = {}
+
+            total_forms = dict(
+                [(ct.__name__.lower(), 0) for ct in self.model._feincms_content_types]
+            )
+
+            for version in revision.version_set.all().select_related("content_type"):
+                if version.object_version.object == obj:
+                    FORM_DATA.update(version.field_dict)
+                    continue
+
+                version_prefix = "%s-%s" % (
+                    version.content_type.model,
+                    total_forms[version.content_type.model]
+                )
+
+
+                for k, v in version.field_dict.items():
+                    form_key = "%s-%s" % (version_prefix, k)
+                    assert form_key not in FORM_DATA
+                    FORM_DATA[form_key] = v
+
+                # defaultdict would be cleaner but this works with Python 2.4:
+                total_forms[version.content_type.model] += 1
+
+
+            for k, v in total_forms.items():
+                FORM_DATA["%s-INITIAL_FORMS" % k] = v
+                # TOTAL FORMS should be one for each actual object and one for
+                # the "Add new" feature. We'll bump the total up if we actually
+                # have existing content:
+                if v:
+                    FORM_DATA["%s-TOTAL_FORMS" % k] = v + 1
+                else:
+                    FORM_DATA["%s-TOTAL_FORMS" % k] = 0
+
+            # BUG: This somehow does not correctly initialize the initial form for adding new content correctly
+            model_form = ModelForm(FORM_DATA, instance=obj)
+            inline_formsets = [
+                formset_class(FORM_DATA, instance=obj, prefix=content_type.__name__.lower())
+                for content_type, formset_class in inline_formset_types
+            ]
         else:
             model_form = ModelForm(instance=obj)
             inline_formsets = [
                 formset_class(instance=obj, prefix=content_type.__name__.lower())
-                for content_type, formset_class in inline_formset_types]
+                for content_type, formset_class in inline_formset_types
+            ]
 
         # Prepare mapping of content types to their prettified names
         content_types = []
@@ -177,6 +247,7 @@ class ItemEditor(admin.ModelAdmin):
             context['available_templates'] = self.model._feincms_templates
 
         context.update({
+            'change': obj.pk is not None,
             'title': _('Change %s') % force_unicode(opts.verbose_name),
             'opts': opts,
             'object': obj,
@@ -189,6 +260,7 @@ class ItemEditor(admin.ModelAdmin):
             'errors': helpers.AdminErrorList(model_form, inline_formsets),
             'FEINCMS_ADMIN_MEDIA': settings.FEINCMS_ADMIN_MEDIA,
             'FEINCMS_ADMIN_MEDIA_HOTLINKING': settings.FEINCMS_ADMIN_MEDIA_HOTLINKING,
+            'messages': messages,
         })
 
         return self.render_item_editor(request, obj, context)
