@@ -1,14 +1,17 @@
 from django.conf import settings as django_settings
 from django.contrib import admin
-from django.contrib.admin.util import unquote
-from django.http import HttpResponse, HttpResponseBadRequest
+from django.db.models.query import QuerySet
+from django.http import HttpResponse, HttpResponseBadRequest, HttpResponseForbidden, HttpResponseNotFound, HttpResponseServerError
 from django.utils import simplejson
 from django.utils.safestring import mark_safe
 from django.utils.translation import ugettext_lazy as _
 
 from feincms import settings
 
+import logging
 
+
+# ------------------------------------------------------------------------
 def django_boolean_icon(field_val, alt_text=None, title=None):
     """
     Return HTML code for a nice representation of true/false.
@@ -39,15 +42,20 @@ def _build_tree_structure(cls):
     """
     all_nodes = { }
     def add_as_descendant(n, p):
-        if not n: return
-        all_nodes[n.id]['descendants'].append(p.id)
-        add_as_descendant(n.parent, p)
+        all_nodes[n]['descendants'].append(p)
 
-    for p in cls.objects.order_by('tree_id', 'lft'):
-        all_nodes[p.id] = { 'id': p.id, 'children' : [ ], 'descendants' : [ ], 'parent' : p.parent_id }
-        if(p.parent_id):
-            all_nodes[p.parent_id]['children'].append(p.id)
-            add_as_descendant(p.parent, p)
+        n_parent_id = all_nodes[n]['parent']
+
+        if n_parent_id:
+            add_as_descendant(n_parent_id, p)
+
+    for p_id, parent_id in cls.objects.order_by(cls._meta.tree_id_attr, cls._meta.left_attr).values_list("pk", "%s_id" % cls._meta.parent_attr):
+
+        all_nodes[p_id] = { 'id': p_id, 'children' : [ ], 'descendants' : [ ], 'parent' : parent_id }
+
+        if parent_id:
+            all_nodes[parent_id]['children'].append(p_id)
+            add_as_descendant(parent_id, p_id)
 
     return all_nodes
 
@@ -107,20 +115,82 @@ def ajax_editable_boolean(attr, short_description):
     return _fn
 
 
-class TreeEditor(admin.ModelAdmin):
-    class Media:
-        css = {}
-        js = []
-        if settings.FEINCMS_ADMIN_MEDIA_HOTLINKING:
-            js.extend(( "http://ajax.googleapis.com/ajax/libs/jquery/1.3.2/jquery.min.js", ))
-        else:
-            js.extend(( settings.FEINCMS_ADMIN_MEDIA + "jquery-1.3.2.min.js", ))
+# ------------------------------------------------------------------------
+class TreeEditorQuerySet(QuerySet):
+    """
+    The TreeEditorQuerySet is a special query set used only in the TreeEditor
+    ChangeList page. The only difference to a regular QuerySet is that it
+    will enforce:
 
-        js.extend(( settings.FEINCMS_ADMIN_MEDIA + "ie_compat.js",
-                    settings.FEINCMS_ADMIN_MEDIA + "jquery.cookie.js" ,
-                    settings.FEINCMS_ADMIN_MEDIA + "toolbox.js",
-                    settings.FEINCMS_ADMIN_MEDIA + "page_toolbox.js",
-                    ))
+        (a) The result is ordered in correct tree order so that
+            the TreeAdmin works all right.
+
+        (b) It ensures that all ancestors of selected items are included
+            in the result set, so the resulting tree display actually
+            makes sense.
+    """
+    def iterator(self):
+        qs = self
+        # Reaching into the bowels of query sets to find out whether the qs is
+        # actually filtered and we need to do the INCLUDE_ANCESTORS dance at all.
+        # INCLUDE_ANCESTORS is quite expensive, so don't do it if not needed.
+        is_filtered = bool(qs.query.where.children)
+        if settings.FEINCMS_TREE_EDITOR_INCLUDE_ANCESTORS and is_filtered:
+            include_pages = set()
+            # Order by 'rght' will return the tree deepest nodes first;
+            # this cuts down the number of queries considerably since all ancestors
+            # will already be in include_pages when they are checked, thus not
+            # trigger additional queries.
+            for p in super(TreeEditorQuerySet, self.order_by('rght')).iterator():
+                if p.parent_id and p.parent_id not in include_pages and \
+                                   p.id not in include_pages:
+                    ancestor_id_list = p.get_ancestors().values_list('id', flat=True)
+                    include_pages.update(ancestor_id_list)
+
+            if include_pages:
+                qs = qs | self.model._default_manager.filter(id__in=include_pages)
+
+            qs = qs.distinct()
+
+        for obj in super(TreeEditorQuerySet, qs).iterator():
+            yield obj
+
+    if settings.FEINCMS_TREE_EDITOR_INCLUDE_ANCESTORS:
+        def __getitem__(self, index):
+            return self   # Don't even try to slice
+
+        def get(self, *args, **kwargs):
+            """
+            Quick and dirty hack to fix change_view and delete_view; they use
+            self.queryset(request).get(...) to get the object they should work
+            with. Our modifications to the queryset when INCLUDE_ANCESTORS is
+            enabled make get() fail often with a MultipleObjectsReturned
+            exception.
+            """
+            return self.model._default_manager.get(*args, **kwargs)
+
+
+# !!!: Hack alert! Patching ChangeList, check whether this still applies post Django 1.1
+# If the ChangeList is used by a TreEditor, we always need to order by 'tree_id' and 'lft'.
+from django.contrib.admin.views import main
+class ChangeList(main.ChangeList):
+    def get_query_set(self):
+        qs = super(ChangeList, self).get_query_set()
+        if isinstance(self.model_admin, TreeEditor):
+            return qs.order_by('tree_id', 'lft')
+        return qs
+main.ChangeList = ChangeList
+
+
+# ------------------------------------------------------------------------
+# MARK: -
+# ------------------------------------------------------------------------
+
+class TreeEditor(admin.ModelAdmin):
+    if settings.FEINCMS_TREE_EDITOR_INCLUDE_ANCESTORS:
+        # Make sure that no pagination is displayed. Slicing is disabled anyway,
+        # therefore this value does not have an influence on the queryset
+        list_per_page = 999999999
 
     def __init__(self, *args, **kwargs):
         super(TreeEditor, self).__init__(*args, **kwargs)
@@ -146,13 +216,23 @@ class TreeEditor(admin.ModelAdmin):
         Generate a short title for a page, indent it depending on
         the page's depth in the hierarchy.
         """
-        r = '''<span onclick="return page_tree_handler('%d')" id="page_marker-%d"
+        if hasattr(item, 'get_absolute_url'):
+            r = '''<input type="hidden" class="medialibrary_file_path" value="%s"><span onclick="return page_tree_handler('%d')" id="page_marker-%d"
+            class="page_marker" style="width: %dpx;">&nbsp;</span>&nbsp;''' % (
+                item.get_absolute_url(),
+                item.id, item.id, 14+item.level*18)
+        else:
+            r = '''<span onclick="return page_tree_handler('%d')" id="page_marker-%d"
             class="page_marker" style="width: %dpx;">&nbsp;</span>&nbsp;''' % (
                 item.id, item.id, 14+item.level*18)
-
+                
+#        r += '<span tabindex="0">'
         if hasattr(item, 'short_title'):
-            return mark_safe(r + item.short_title())
-        return mark_safe(r + unicode(item))
+            r += item.short_title()
+        else:
+            r += unicode(item)
+#        r += '</span>'
+        return mark_safe(r)
     indented_short_title.short_description = _('title')
     indented_short_title.allow_tags = True
 
@@ -196,19 +276,40 @@ class TreeEditor(admin.ModelAdmin):
         """
         Handle an AJAX toggle_boolean request
         """
-        self._collect_editable_booleans()
+        try:
+            item_id = int(request.POST.get('item_id', None))
+            attr = str(request.POST.get('attr', None))
+        except:
+            return HttpResponseBadRequest("Malformed request")
 
-        item_id = request.POST.get('item_id')
-        attr = request.POST.get('attr')
+        if not request.user.is_staff:
+            logging.warning("Denied AJAX request by non-staff %s to toggle boolean %s for page #%s", request.user, attr, item_id)
+            return HttpResponseForbidden("You do not have permission to access this page")
+
+        self._collect_editable_booleans()
 
         if not self._ajax_editable_booleans.has_key(attr):
             return HttpResponseBadRequest("not a valid attribute %s" % attr)
 
         try:
-            obj = self.model._default_manager.get(pk=unquote(item_id))
+            obj = self.model._default_manager.get(pk=item_id)
+        except self.model.DoesNotExist:
+            return HttpResponseNotFound("Object does not exist")
 
-            attr = str(attr)
+        can_change = False
 
+        if hasattr(obj, "user_can") and obj.user_can(request.user, change_page=True):
+            can_change = True
+        else:
+            can_change = request.user.has_perm("page.change_page")
+
+        if not can_change:
+            logging.warning("Denied AJAX request by %s to toggle boolean %s for page %s", request.user, attr, item_id)
+            return HttpResponseForbidden("You do not have permission to access this page")
+
+        logging.info("Processing request by %s to toggle %s on %s", request.user, attr, obj)
+
+        try:
             before_data = self._ajax_editable_booleans[attr](self, obj)
 
             setattr(obj, attr, not getattr(obj, attr))
@@ -220,7 +321,8 @@ class TreeEditor(admin.ModelAdmin):
             data = self._ajax_editable_booleans[attr](self, obj)
 
         except Exception, e:
-            return HttpResponse("FAILED " + unicode(e), mimetype="text/plain")
+            logging.exception("Unhandled exception while toggling %s on %s", attr, obj)
+            return HttpResponseServerError("Unable to toggle %s on %s" % (attr, obj))
 
         # Weed out unchanged cells to keep the updates small. This assumes
         # that the order a possible get_descendents() returns does not change
@@ -249,12 +351,13 @@ class TreeEditor(admin.ModelAdmin):
             elif cmd == 'move_node':
                 return self._move_node(request)
             else:
-                return HttpResponse('Oops. AJAX request not understood.')
+                return HttpResponseBadRequest('Oops. AJAX request not understood.')
 
         self._refresh_changelist_caches()
 
         extra_context = extra_context or {}
         extra_context['FEINCMS_ADMIN_MEDIA'] = settings.FEINCMS_ADMIN_MEDIA
+        extra_context['FEINCMS_ADMIN_MEDIA_HOTLINKING'] = settings.FEINCMS_ADMIN_MEDIA_HOTLINKING
         extra_context['tree_structure'] = mark_safe(simplejson.dumps(
                                                     _build_tree_structure(self.model)))
 
@@ -280,13 +383,13 @@ class TreeEditor(admin.ModelAdmin):
         Returns a QuerySet of all model instances that can be edited by the
         admin site. This is used by changelist_view.
         """
-
-        # Use default ordering, always
-        return self.model._default_manager.get_query_set()
+        qs = self.model._default_manager.get_query_set()
+        qs.__class__ = TreeEditorQuerySet
+        return qs
 
     def _actions_column(self, page):
         actions = []
-        actions.append(u'<a href="#" onclick="return cut_item(\'%s\', this)" title="%s">&#x2702;</a>' % (
+        actions.append(u'<a href="#" onclick="return cut_item(\'%s\', this)" title="%s"><big>&#x2702;</big></a>' % (
             page.pk, _('Cut')))
 
         actions.append(u'<a class="paste_target" href="#" onclick="return paste_item(\'%s\', \'last-child\')" title="%s">&#x21b3;</a>' % (
@@ -300,18 +403,3 @@ class TreeEditor(admin.ModelAdmin):
     actions_column.allow_tags = True
     actions_column.short_description = _('actions')
 
-
-# !!!: Hack alert! Patching ChangeList, check whether this still applies post Django 1.1
-# Note: Patch lifted from http://code.djangoproject.com/ticket/4926 (ticket_4926_changelist_multifield_ordering.diff)
-
-# By default, django only orders by first ordering field in admin. We patch
-# up the ChangeList here so it returns "use default ordering" for any Page
-# lookups. That way, we can order by tree_id + lft and get the site's natural
-# page structure.
-from django.contrib.admin.views import main
-class ChangeList(main.ChangeList):
-    def get_ordering(self):
-        if isinstance(self.model_admin, TreeEditor):
-            return '', ''
-        return super(ChangeList, self).get_ordering()
-main.ChangeList = ChangeList
