@@ -5,13 +5,16 @@ All models defined here are abstract, which means no tables are created in
 the feincms_ namespace.
 """
 
+from django.contrib.contenttypes.models import ContentType
 from django.core.exceptions import ImproperlyConfigured
 from django.db import models
+from django.db.models import Q
 from django.template.loader import render_to_string
 from django.utils.encoding import force_unicode
 from django.utils.translation import ugettext_lazy as _
 
-from feincms.utils import copy_model_instance
+from feincms import settings
+from feincms.utils import get_object, copy_model_instance
 
 try:
     any
@@ -82,6 +85,8 @@ class Base(models.Model):
     class Meta:
         abstract = True
 
+    _cached_django_content_type = None
+
     @classmethod
     def register_regions(cls, *regions):
         """
@@ -136,7 +141,7 @@ class Base(models.Model):
 
         cls.TEMPLATE_CHOICES = choices
 
-        cls.add_to_class('template_key', models.CharField(_('template'), max_length=20,
+        cls.add_to_class('template_key', models.CharField(_('template'), max_length=255,
             choices=cls.TEMPLATE_CHOICES, default=choices[0][0]))
 
         cls._feincms_templates = instances
@@ -149,6 +154,43 @@ class Base(models.Model):
         for template in cls._feincms_templates.values():
             cls._feincms_all_regions += template.regions
 
+    @classmethod
+    def register_extension(cls, register_fn):
+        """
+        Call the register function of an extension. You must override this
+        if you provide a custom ModelAdmin class and want your extensions to
+        be able to patch stuff in.
+        """
+        register_fn(cls, None)
+
+    @classmethod
+    def register_extensions(cls, *extensions):
+        if not hasattr(cls, '_feincms_extensions'):
+            cls._feincms_extensions = set()
+
+        here = cls.__module__.split('.')[:-1]
+        here_path = '.'.join(here + ['extensions'])
+        common_path = '.'.join(here[:-1] + ['extensions'])
+
+        for ext in extensions:
+            if ext in cls._feincms_extensions:
+                continue
+
+            try:
+                try:
+                    fn = get_object(ext + '.register')
+                except ImportError:
+                    try:
+                        fn = get_object('%s.%s.register' % ( here_path, ext ) )
+                    except ImportError:
+                        fn = get_object('%s.%s.register' % ( common_path, ext ) )
+
+                cls.register_extension(fn)
+                cls._feincms_extensions.add(ext)
+            except Exception, e:
+                raise ImproperlyConfigured("%s.register_extensions('%s') raised an '%s' exception" %
+                                            (cls.__name__, ext, e.message))
+
     @property
     def content(self):
         """
@@ -160,27 +202,48 @@ class Base(models.Model):
 
         return self._content_proxy
 
+    def _get_content_types_for_region(self, region):
+        # find all concrete content type tables which have at least one entry for
+        # the current CMS object and region
+        # This method is overridden by a more efficient implementation if
+        # the ct_tracker extension is active.
+
+        from django.core.cache import cache as django_cache
+
+        counts = None
+        ck = None
+        # ???: Should we move the cache_key() method to Base, so we can avoid
+        # the if-it-supports-it dance?
+        if settings.FEINCMS_USE_CACHE and getattr(self, 'cache_key', None):
+            ck = 'CNT-FOR-REGION-' + region.key + '-' + self.cache_key()
+            counts = django_cache.get(ck)
+
+        if counts is None:
+            sql = ' UNION '.join([
+                'SELECT %d AS ct_idx, COUNT(id) FROM %s WHERE parent_id=%s AND region=%%s' % (
+                    idx,
+                    cls._meta.db_table,
+                    self.pk) for idx, cls in enumerate(self._feincms_content_types)])
+            sql = 'SELECT * FROM ( ' + sql + ' ) AS ct ORDER BY ct_idx'
+
+            from django.db import connection
+            cursor = connection.cursor()
+            cursor.execute(sql, [region.key] * len(self._feincms_content_types))
+
+            counts = [row[1] for row in cursor.fetchall()]
+
+            if ck:
+                django_cache.set(ck, counts)
+
+        return counts
+
     def _content_for_region(self, region):
         """
         This method is used primarily by the ContentProxy
         """
-
         self._needs_content_types()
 
-        # find all concrete content type tables which have at least one entry for
-        # the current CMS object and region
-        sql = ' UNION '.join([
-            'SELECT %d AS ct_idx, COUNT(id) FROM %s WHERE parent_id=%s AND region=%%s' % (
-                idx,
-                cls._meta.db_table,
-                self.pk) for idx, cls in enumerate(self._feincms_content_types)])
-        sql = 'SELECT * FROM ( ' + sql + ' ) AS ct ORDER BY ct_idx'
-
-        from django.db import connection
-        cursor = connection.cursor()
-        cursor.execute(sql, [region.key] * len(self._feincms_content_types))
-
-        counts = [row[1] for row in cursor.fetchall()]
+        counts = self._get_content_types_for_region(region)
 
         if not any(counts):
             return []
@@ -192,9 +255,8 @@ class Base(models.Model):
                 # of different type will have to be sorted into a list according
                 # to their 'ordering' attribute later
                 contents += list(
-                    self._feincms_content_types[idx].objects.filter(
-                        parent=self,
-                        region=region.key).select_related())
+                    self._feincms_content_types[idx].get_queryset(
+                        Q(parent=self) & Q(region=region.key)))
 
         return contents
 
@@ -267,6 +329,9 @@ class Base(models.Model):
                 self.id,
                 )
 
+        def get_queryset(cls, filter_args):
+            return cls.objects.select_related().filter(filter_args)
+
         attrs = {
             '__module__': cls.__module__, # The basic content type is put into
                                           # the same module as the CMS base type.
@@ -278,9 +343,10 @@ class Base(models.Model):
             'render': render,
             'fe_render': fe_render,
             'fe_identifier': fe_identifier,
+            'get_queryset': classmethod(get_queryset),
             'Meta': Meta,
             'parent': models.ForeignKey(cls, related_name='%(class)s_set'),
-            'region': models.CharField(max_length=20),
+            'region': models.CharField(max_length=255),
             'ordering': models.IntegerField(_('ordering'), default=0),
             }
 
@@ -298,11 +364,6 @@ class Base(models.Model):
         # by content types
         cls.feincms_item_editor_includes = {}
 
-        # Fetch django's content type
-        from django.contrib.contenttypes.models import ContentType
-        ct = ContentType.objects.get_for_model(cls)
-        cls._content_type = ct
-
     @classmethod
     def create_content_type(cls, model, regions=None, **kwargs):
         """
@@ -327,10 +388,14 @@ class Base(models.Model):
         # with the same class name because of related_name clashes
         try:
             getattr(cls, '%s_set' % model.__name__.lower())
-            raise ImproperlyConfigured, 'Cannot create content type using %s.%s for %s.%s, because %s_set is already taken.' % (
-                model.__module__, model.__name__,
-                cls.__module__, cls.__name__,
-                model.__name__.lower())
+            import warnings
+            warnings.warn(
+                'Cannot create content type using %s.%s for %s.%s, because %s_set is already taken.' % (
+                    model.__module__, model.__name__,
+                    cls.__module__, cls.__name__,
+                    model.__name__.lower()),
+                RuntimeWarning)
+            return
         except AttributeError:
             # everything ok
             pass
@@ -343,7 +408,7 @@ class Base(models.Model):
 
         feincms_content_base = cls._feincms_content_model
 
-        class Meta:
+        class Meta(feincms_content_base.Meta):
             db_table = '%s_%s' % (cls._meta.db_table, model.__name__.lower())
             verbose_name = model._meta.verbose_name
             verbose_name_plural = model._meta.verbose_name_plural
@@ -382,6 +447,9 @@ class Base(models.Model):
 
         model._feincms_content_models.append(new_type)
 
+        # Add a backlink from content-type to content holder class
+        new_type._feincms_content_class = cls
+
         # customization hook.
         if hasattr(new_type, 'initialize_type'):
             new_type.initialize_type(**kwargs)
@@ -397,10 +465,16 @@ class Base(models.Model):
         # collect item editor includes from the content type
         if hasattr(model, 'feincms_item_editor_includes'):
             for key, includes in model.feincms_item_editor_includes.items():
-                cls.feincms_item_editor_includes.setdefault(key, []).extend(includes)
+                cls.feincms_item_editor_includes.setdefault(key, set()).update(includes)
 
         return new_type
 
+    @property
+    def _django_content_type(self):
+        if getattr(self.__class__, '_cached_django_content_type', None) is None:
+            self.__class__._cached_django_content_type = ContentType.objects.get_for_model(self)
+        return self.__class__._cached_django_content_type
+
     @classmethod
     def content_type_for(cls, model):
         """
@@ -423,7 +497,9 @@ class Base(models.Model):
         # helper which can be used to ensure that either register_regions or
         # register_templates has been executed before proceeding
         if not hasattr(cls, 'template'):
-            raise ImproperlyConfigured, 'You need to register at least one template for Page before the admin code is included.'
+            raise ImproperlyConfigured, 'You need to register at least one template or one region on %s.' % (
+                cls.__name__,
+                )
 
     @classmethod
     def _needs_content_types(cls):
@@ -470,6 +546,8 @@ class ContentProxy(object):
         has the inherited flag set, this method will go up the ancestor chain
         until either some item contents have found or no ancestors are left.
         """
+        if (attr.startswith('__')):
+            raise AttributeError
 
         item = self.__dict__['item']
 
@@ -482,9 +560,9 @@ class ContentProxy(object):
             contents = obj._content_for_region(region)
 
             # go to parent if this model has a parent attribute
-            # TODO this should be abstracted into a property/method or something
+            # TODO: this should be abstracted into a property/method or something
             # The link which should be followed is not always '.parent'
-            if not contents and hasattr(obj, 'parent_id') and obj.parent_id and region.inherited:
+            if region.inherited and not contents and hasattr(obj, 'parent_id') and obj.parent_id:
                 return collect_items(obj.parent)
 
             return contents
