@@ -15,7 +15,7 @@ class EntryManager(models.Manager):
     def published(self):
         return self.filter(
             published__isnull=False,
-            published__lte=datetime.now(),
+            published_on__lte=datetime.now(),
             )
 
 
@@ -29,6 +29,7 @@ class Entry(Base):
         help_text=_('Will be set automatically once you tick the `published` checkbox above.'))
 
     class Meta:
+        get_latest_by = 'published_on'
         ordering = ['-published_on']
         verbose_name = _('entry')
         verbose_name_plural = _('entries')
@@ -48,17 +49,9 @@ class Entry(Base):
         return ('blog_entry_detail', (self.id,), {})
 
     @classmethod
-    def register_extensions(cls, *extensions):
-        if not hasattr(cls, '_feincms_extensions'):
-            cls._feincms_extensions = set()
+    def register_extension(cls, register_fn):
+        register_fn(cls, EntryAdmin)
 
-        for ext in extensions:
-            if ext in cls._feincms_extensions:
-                continue
-
-            fn = get_object('feincms.module.blog.extensions.%s.register' % ext)
-            fn(cls, EntryAdmin)
-            cls._feincms_extensions.add(ext)
 
 signals.post_syncdb.connect(check_database_schema(Entry, __name__), weak=False)
 
