@@ -1,6 +1,5 @@
 # ------------------------------------------------------------------------
 # coding=utf-8
-# $Id$
 # ------------------------------------------------------------------------
 
 from django import forms
@@ -13,7 +12,8 @@ from django.forms.models import model_to_dict
 from django.forms.util import ErrorList
 from django.http import Http404, HttpResponseRedirect
 from django.utils.safestring import mark_safe
-from django.utils.translation import ugettext_lazy as _
+from django.utils.translation import ugettext_lazy as _, ugettext
+from django.db.transaction import commit_on_success
 
 import mptt
 
@@ -22,8 +22,10 @@ from feincms.admin import editor
 from feincms.management.checker import check_database_schema
 from feincms.models import Base
 from feincms.utils import get_object, copy_model_instance
+import feincms.admin.filterspecs
 
 
+# ------------------------------------------------------------------------
 class PageManager(models.Manager):
 
     # A list of filters which are used to determine whether a page is active or not.
@@ -89,18 +91,37 @@ class PageManager(models.Manager):
         paths = ['/']
         path = path.strip('/')
 
+        # Cache path -> page resolving.
+        # Note: The only possibility that this returns a stale association
+        # is if a page at an url is removed and/or replaced by another page
+        # at that url. This might happen either by manual intervention (admin
+        # moving the page) or automatically via the datepublisher extension
+        # (a page expiring and another becoming active).
+        # Both cases can handle a short time of "being wrong", so this should
+        # be OK.
+
+        from django.core.cache import cache as django_cache
+        if settings.FEINCMS_USE_CACHE:
+            ck = 'PAGE-FOR-URL-' + path
+            page = django_cache.get(ck)
+            if page:
+                return page
+
         if path:
             tokens = path.split('/')
             paths += ['/%s/' % '/'.join(tokens[:i]) for i in range(1, len(tokens)+1)]
 
         try:
-            return self.active().filter(_cached_url__in=paths).extra(
+            page = self.active().filter(_cached_url__in=paths).extra(
                 select={'_url_length': 'LENGTH(_cached_url)'}).order_by('-_url_length')[0]
+            if settings.FEINCMS_USE_CACHE:
+                django_cache.set(ck, page)
+            return page
         except IndexError:
             if raise404:
                 raise Http404
 
-        return None
+        raise self.model.DoesNotExist
 
     def in_navigation(self):
         return self.active().filter(in_navigation=True)
@@ -162,24 +183,26 @@ class PageManager(models.Manager):
 # MARK: -
 # ------------------------------------------------------------------------
 class Page(Base):
+
     active = models.BooleanField(_('active'), default=False)
 
     # structure and navigation
-    title = models.CharField(_('title'), max_length=100,
+    title = models.CharField(_('title'), max_length=200,
         help_text=_('This is used for the generated navigation too.'))
-    slug = models.SlugField(_('slug'))
+    slug = models.SlugField(_('slug'), max_length=150)
     parent = models.ForeignKey('self', blank=True, null=True, related_name='children')
+    parent.parent_filter = True # Custom list_filter - see admin/filterspecs.py
     in_navigation = models.BooleanField(_('in navigation'), default=True)
-    override_url = models.CharField(_('override URL'), max_length=200, blank=True,
+    override_url = models.CharField(_('override URL'), max_length=400, blank=True,
         help_text=_('Override the target URL. Be sure to include slashes at the beginning and at the end if it is a local URL. This affects both the navigation and subpages\' URLs.'))
-    redirect_to = models.CharField(_('redirect to'), max_length=200, blank=True,
+    redirect_to = models.CharField(_('redirect to'), max_length=400, blank=True,
         help_text=_('Target URL for automatic redirects.'))
-    _cached_url = models.CharField(_('Cached URL'), max_length=200, blank=True,
+    _cached_url = models.CharField(_('Cached URL'), max_length=400, blank=True,
         editable=False, default='', db_index=True)
 
     request_processors = []
     response_processors = []
-    cache_key_components = [ lambda p: p._content_type.id,
+    cache_key_components = [ lambda p: p._django_content_type.id,
                              lambda p: p.id ]
 
     class Meta:
@@ -190,7 +213,16 @@ class Page(Base):
     objects = PageManager()
 
     def __unicode__(self):
-        return mark_safe('&nbsp;&nbsp;' * self.level + self.short_title())
+        return self.short_title()
+
+    def is_active(self):
+        """
+        Check whether this page and all its ancestors are active
+        """
+
+        pages = Page.objects.active().filter(tree_id=self.tree_id, lft__lte=self.lft, rght__gte=self.rght)
+        return pages.count() > self.level
+    is_active.short_description = _('is active')
 
     def are_ancestors_active(self):
         """
@@ -205,21 +237,20 @@ class Page(Base):
 
     def short_title(self):
         """
-        Do a short version of the title, truncate it intelligently when too long.
-        Try to cut it in 2/3 + ellipsis + 1/3 of the original title. The first part
-        also tries to cut at white space instead of in mid-word.
+        Title shortened for display.
         """
-        max_length = 50
-        if len(self.title) >= max_length:
-            first_part = int(max_length * 0.6)
-            next_space = self.title[first_part:(max_length / 2 - first_part)].find(' ')
-            if next_space >= 0:
-                first_part += next_space
-            return self.title[:first_part] + u' … ' + self.title[-(max_length - first_part):]
-        return self.title
+        from feincms.utils import shorten_string
+        return shorten_string(self.title)
     short_title.admin_order_field = 'title'
     short_title.short_description = _('title')
 
+    def __init__(self, *args, **kwargs):
+        super(Page, self).__init__(*args, **kwargs)
+        # Cache a copy of the loaded _cached_url value so we can reliably
+        # determine whether it has been changed in the save handler:
+        self._original_cached_url = self._cached_url
+
+    @commit_on_success
     def save(self, *args, **kwargs):
         cached_page_urls = {}
 
@@ -234,8 +265,15 @@ class Page(Base):
         cached_page_urls[self.id] = self._cached_url
         super(Page, self).save(*args, **kwargs)
 
-        # make sure that we get the descendants back after their parents
+        # If our cached URL changed we need to update all descendants to
+        # reflect the changes. Since this is a very expensive operation
+        # on large sites we'll check whether our _cached_url actually changed
+        # or if the updates weren't navigation related:
+        if self._cached_url == self._original_cached_url:
+            return
+
         pages = self.get_descendants().order_by('lft')
+
         for page in pages:
             if page.override_url:
                 page._cached_url = page.override_url
@@ -260,6 +298,9 @@ class Page(Base):
     def get_navigation_url(self):
         return self.redirect_to or self._cached_url
 
+    def get_siblings_and_self(page):
+        return page.get_siblings(include_self=True)
+
     def cache_key(self):
         """
         Return a string that may be used as cache key for the current page.
@@ -309,13 +350,31 @@ class Page(Base):
         if not self.are_ancestors_active():
             raise Http404()
 
+    def get_redirect_to_target(self, request):
+        """
+        This might be overriden/extended by extension modules.
+        """
+        return self.redirect_to
+
     def redirect_request_processor(self, request):
-        if self.redirect_to:
-            return HttpResponseRedirect(self.redirect_to)
+        target = self.get_redirect_to_target(request)
+        if target:
+            return HttpResponseRedirect(target)
 
     def frontendediting_request_processor(self, request):
-        if 'frontend_editing' in request.GET and request.user.has_module_perms('page'):
-            request.session['frontend_editing'] = request.GET['frontend_editing'] and True or False
+        if not 'frontend_editing' in request.GET:
+            return
+
+        if request.user.has_module_perms('page'):
+            try:
+                enable_fe = int(request.GET['frontend_editing']) > 0
+            except ValueError:
+                enable_fe = False
+
+            request.session['frontend_editing'] = enable_fe
+
+        # Redirect to cleanup URLs
+        return HttpResponseRedirect(request.path)
 
     def etag_request_processor(self, request):
 
@@ -388,7 +447,7 @@ class Page(Base):
                 time += float(q['time'])
 
             print "--------------------------------------------------------------"
-            print "Total: %d queries, %f ms" % (i, time)
+            print "Total: %d queries, %.3f ms" % (i, time)
             print "--------------------------------------------------------------"
 
         return processor
@@ -402,21 +461,9 @@ class Page(Base):
         cls.response_processors.extend(processors)
 
     @classmethod
-    def register_extensions(cls, *extensions):
-        if not hasattr(cls, '_feincms_extensions'):
-            cls._feincms_extensions = set()
+    def register_extension(cls, register_fn):
+        register_fn(cls, PageAdmin)
 
-        for ext in extensions:
-            if ext in cls._feincms_extensions:
-                continue
-
-            try:
-                fn = get_object(ext + '.register')
-            except ImportError:
-                fn = get_object('feincms.module.page.extensions.%s.register' % ext)
-
-            fn(cls, PageAdmin)
-            cls._feincms_extensions.add(ext)
 
 # ------------------------------------------------------------------------
 mptt.register(Page)
@@ -444,7 +491,7 @@ class PageAdminForm(forms.ModelForm):
                             del data[field]
 
                     # These are always excluded from prefilling
-                    for field in ('title', 'slug', 'parent', 'active'):
+                    for field in ('title', 'slug', 'parent', 'active', 'override_url'):
                         del data[field]
 
                     kwargs['initial'].update(data)
@@ -457,7 +504,12 @@ class PageAdminForm(forms.ModelForm):
                     page = Page.objects.get(pk=kwargs['initial']['translation_of'])
                     original = page.original_translation
 
-                    data = {'translation_of': original.id, 'template_key': original.template_key}
+                    data = {
+                        'translation_of': original.id,
+                        'template_key': original.template_key,
+                        'active': original.active,
+                        'in_navigation': original.in_navigation,
+                        }
 
                     if original.parent:
                         try:
@@ -485,7 +537,11 @@ class PageAdminForm(forms.ModelForm):
             self.fields['template_key'].choices = choices
 
     def clean(self):
-        cleaned_data = self.cleaned_data
+        cleaned_data = super(PageAdminForm, self).clean()
+
+        # No need to think further, let the user correct errors first
+        if self._errors:
+            return cleaned_data
 
         current_id = None
         # See the comment below on why we do not use Page.objects.active(),
@@ -535,10 +591,25 @@ if settings.FEINCMS_PAGE_USE_SPLIT_PANE_EDITOR:
 else:
     list_modeladmin = editor.TreeEditor
 
-
 # MARK: -
 # ------------------------------------------------------------------------
+
+# We'll create a replacement for the normal "delete_selected" action which is
+# slower, calling .delete() on each object rather than the queryset .delete()
+# but avoids problems with hierarchy leading to all pages being deleted.
+#
+# See the warning here:
+# http://docs.djangoproject.com/en/dev/ref/contrib/admin/actions/ for details
+def delete_page_action(modeladmin, request, queryset):
+    for i in queryset:
+        i.delete()
+delete_page_action.short_description = "Delete selected pages"
+
 class PageAdmin(editor.ItemEditor, list_modeladmin):
+    class Media:
+        css = {}
+        js = []
+
     form = PageAdminForm
 
     # the fieldsets config here is used for the add_view, it has no effect
@@ -558,12 +629,25 @@ class PageAdmin(editor.ItemEditor, list_modeladmin):
     search_fields = ['title', 'slug']
     prepopulated_fields = { 'slug': ('title',), }
 
-    raw_id_fields = []
-    show_on_top = ['title', 'active']
+    raw_id_fields = ['parent']
+    show_on_top = ['title', 'active', 'parent']
     radio_fields = {'template_key': admin.HORIZONTAL}
 
+    actions = [ delete_page_action ]
+
+    def __init__(self, *args, **kwargs):
+        if len(Page._feincms_templates) > 4:
+            del(self.radio_fields['template_key'])
+
+        return super(PageAdmin, self).__init__(*args, **kwargs)
+
     in_navigation_toggle = editor.ajax_editable_boolean('in_navigation', _('in navigation'))
 
+    def get_actions(self, request):
+        actions = super(PageAdmin, self).get_actions(request)
+        del actions['delete_selected']
+        return actions
+
     def _actions_column(self, page):
         actions = super(PageAdmin, self)._actions_column(page)
         actions.insert(0, u'<a href="add/?parent=%s" title="%s"><img src="%simg/admin/icon_addlink.gif" alt="%s"></a>' % (
@@ -574,7 +658,10 @@ class PageAdmin(editor.ItemEditor, list_modeladmin):
 
     def add_view(self, request, form_url='', extra_context=None):
         # Preserve GET parameters
-        return super(PageAdmin, self).add_view(request, request.get_full_path(), extra_context)
+        return super(PageAdmin, self).add_view(
+            request=request,
+            form_url=request.get_full_path(),
+            extra_context=extra_context)
 
     def response_add(self, request, obj, *args, **kwargs):
         response = super(PageAdmin, self).response_add(request, obj, *args, **kwargs)
@@ -596,12 +683,40 @@ class PageAdmin(editor.ItemEditor, list_modeladmin):
     def _refresh_changelist_caches(self, *args, **kwargs):
         self._visible_pages = list(self.model.objects.active().values_list('id', flat=True))
 
-    # ---------------------------------------------------------------------
+    def change_view(self, request, object_id, extra_context=None):
+        from django.shortcuts import get_object_or_404
+        if 'create_copy' in request.GET:
+            page = get_object_or_404(Page, pk=object_id)
+            new = Page.objects.create_copy(page)
+            self.message_user(request, ugettext("You may edit the copied page below."))
+            return HttpResponseRedirect('../%s/' % new.pk)
+        elif 'replace' in request.GET:
+            page = get_object_or_404(Page, pk=request.GET.get('replace'))
+            with_page = get_object_or_404(Page, pk=object_id)
+            Page.objects.replace(page, with_page)
+            self.message_user(request, ugettext("You have replaced %s. You may continue editing the now-active page below.") % page)
+            return HttpResponseRedirect('.')
+
+        return super(PageAdmin, self).change_view(request, object_id, extra_context)
+
+    def render_item_editor(self, request, object, context):
+        try:
+            active = Page.objects.active().exclude(pk=object.pk).get(_cached_url=object._cached_url)
+            context['to_replace'] = active
+        except Page.DoesNotExist:
+            pass
+
+        return super(PageAdmin, self).render_item_editor(request, object, context)
+
     def is_visible_admin(self, page):
         """
         Instead of just showing an on/off boolean, also indicate whether this
         page is not visible because of publishing dates or inherited status.
         """
+        if not hasattr(self, "_visible_pages"):
+            self._visible_pages = list() # Sanity check in case this is not already defined
+
+        # TODO: Couldn't this simply check page.get_ancestors(visible=False).count()?
         if page.parent_id and not page.parent_id in self._visible_pages:
             # parent page's invisibility is inherited
             if page.id in self._visible_pages:
@@ -614,7 +729,7 @@ class PageAdmin(editor.ItemEditor, list_modeladmin):
 
         return editor.ajax_editable_boolean_cell(page, 'active')
     is_visible_admin.allow_tags = True
-    is_visible_admin.short_description = _('is visible')
+    is_visible_admin.short_description = _('is active')
     is_visible_admin.editable_boolean_field = 'active'
 
     # active toggle needs more sophisticated result function
@@ -622,6 +737,8 @@ class PageAdmin(editor.ItemEditor, list_modeladmin):
         retval = []
         for c in page.get_descendants(include_self=True):
             retval.append(self.is_visible_admin(c))
-        return map(lambda page: self.is_visible_admin(page), page.get_descendants(include_self=True))
+        return retval
     is_visible_admin.editable_boolean_result = is_visible_recursive
 
+# ------------------------------------------------------------------------
+# ------------------------------------------------------------------------
