@@ -3,16 +3,21 @@ from django.conf import settings
 from django.http import HttpRequest
 
 from feincms.module.page.models import Page, PageManager
-from feincms.templatetags.utils import *
+from feincms.utils.templatetags import *
+from feincms.utils.templatetags import _parse_args
+
 register = template.Library()
 
 
+# ------------------------------------------------------------------------
+# ------------------------------------------------------------------------
 class NavigationNode(SimpleAssignmentNodeWithVarAndArgs):
     """
     Return a list of pages to be used for the navigation
 
     level: 1 = toplevel, 2 = sublevel, 3 = sub-sublevel
     depth: 1 = only one level, 2 = subpages too
+    extended: run navigation extension on returned pages, not only on top-level node
 
     If you set depth to something else than 1, you might want to look into
     the tree_info template tag from the mptt_tags library.
@@ -31,6 +36,22 @@ class NavigationNode(SimpleAssignmentNodeWithVarAndArgs):
         if isinstance(instance, HttpRequest):
             instance = Page.objects.from_request(instance)
 
+        entries = self._what(instance, level, depth)
+
+        if args.get('extended', False):
+            _entries = list(entries)
+            entries = []
+
+            for entry in _entries:
+                entries.append(entry)
+
+                if getattr(entry, 'navigation_extension', None):
+                    entries.extend(entry.extended_navigation(depth=depth,
+                        request=self.render_context.get('request', None)))
+
+        return entries
+
+    def _what(self, instance, level, depth):
         if level <= 1:
             if depth == 1:
                 return Page.objects.toplevel_navigation()
@@ -62,7 +83,21 @@ class NavigationNode(SimpleAssignmentNodeWithVarAndArgs):
                 return PageManager.apply_active_filters(queryset)
 register.tag('feincms_navigation', do_simple_assignment_node_with_var_and_args_helper(NavigationNode))
 
+class ExtendedNavigationNode(NavigationNode):
+    def render(self, context):
+        self.render_context = context
+        try:
+            instance = self.in_var.resolve(context)
+        except template.VariableDoesNotExist:
+            context[self.var_name] = []
+            return ''
 
+        context[self.var_name] = self.what(instance, _parse_args(self.args, context))
+
+        return ''
+register.tag('feincms_navigation_extended', do_simple_assignment_node_with_var_and_args_helper(ExtendedNavigationNode))
+
+# ------------------------------------------------------------------------
 class ParentLinkNode(SimpleNodeWithVarAndArgs):
     """
     {% feincms_parentlink of feincms_page level=1 %}
@@ -82,7 +117,7 @@ class ParentLinkNode(SimpleNodeWithVarAndArgs):
             return '#'
 register.tag('feincms_parentlink', do_simple_node_with_var_and_args_helper(ParentLinkNode))
 
-
+# ------------------------------------------------------------------------
 class LanguageLinksNode(SimpleAssignmentNodeWithVarAndArgs):
     """
     {% feincms_languagelinks for feincms_page as links [args] %}
@@ -125,48 +160,92 @@ class LanguageLinksNode(SimpleAssignmentNodeWithVarAndArgs):
         return links
 register.tag('feincms_languagelinks', do_simple_assignment_node_with_var_and_args_helper(LanguageLinksNode))
 
+# ------------------------------------------------------------------------
+# ------------------------------------------------------------------------
+def _translate_page_into(page, language, default=None):
+    """
+    Return the translation for a given page
+    """
+    # Optimisation shortcut: No need to dive into translations if page already what we want
+    if page.language == language:
+        return page
 
+    translations = dict((t.language, t) for t in page.available_translations())
+    translations[page.language] = page
+
+    if language in translations:
+        return translations[language]
+    else:
+        if hasattr(default, '__call__'):
+            return default(page=page)
+        return default
+
+# ------------------------------------------------------------------------
 class TranslatedPageNode(SimpleAssignmentNodeWithVarAndArgs):
     """
     {% feincms_translatedpage for feincms_page as feincms_transpage language=en %}
     {% feincms_translatedpage for feincms_page as originalpage %}
+    {% feincms_translatedpage for some_page as translatedpage language=feincms_page.language %}
 
     This template tag needs the translations extension.
 
     Returns the requested translation of the page if it exists. If the language
     argument is omitted the primary language will be returned (the first language
-    specified in settings.LANGUAGES)
+    specified in settings.LANGUAGES).
+
+    Note: To distinguish between a bare language code and a variable we check whether
+    settings LANGUAGES contains that code -- so naming a variable "en" will probably
+    not do what is intended.
     """
-    def what(self, page, args):
+    def what(self, page, args, default=None):
         language = args.get('language',False)
         if not language:
             language = settings.LANGUAGES[0][0]
+        elif language not in (x[0] for x in settings.LANGUAGES):
+            language = template.Variable(language).resolve(self.render_context)
 
-        translations = dict((t.language, t) for t in page.available_translations())
-        translations[page.language] = page
-
-        if language in translations:
-            return translations[language]
-        else:
-            return None
+        return _translate_page_into(page, language, default=default)
 register.tag('feincms_translatedpage', do_simple_assignment_node_with_var_and_args_helper(TranslatedPageNode))
 
+# ------------------------------------------------------------------------
+class TranslatedPageNodeOrBase(TranslatedPageNode):
+    def what(self, page, args):
+        return super(TranslatedPageNodeOrBase, self).what(page, args, default=page.get_original_translation)
+register.tag('feincms_translatedpage_or_base', do_simple_assignment_node_with_var_and_args_helper(TranslatedPageNodeOrBase))
 
-@register.simple_tag
-def feincms_breadcrumbs(page):
+# ------------------------------------------------------------------------
+@register.filter
+def feincms_translated_or_base(pages, language=None):
+    if not hasattr(pages, '__iter__'):
+        pages = [ pages ]
+    for page in pages:
+        yield _translate_page_into(page, language, default=page.get_original_translation)
+
+# ------------------------------------------------------------------------
+@register.inclusion_tag("breadcrumbs.html")
+def feincms_breadcrumbs(page, include_self=True):
     """
+    Generate a list of the page's ancestors suitable for use as breadcrumb navigation.
+
+    By default, generates an unordered list with the id "breadcrumbs" -
+    override breadcrumbs.html to change this.
+
     {% feincms_breadcrumbs feincms_page %}
     """
+
+    if not page or not isinstance(page, Page):
+        raise ValueError("feincms_breadcrumbs must be called with a valid Page object")
+
     ancs = page.get_ancestors()
-    if not ancs:
-        return ""
-    bc = []
-    for anc in ancs:
-        bc.append('<a href="%s">%s</a> &gt; ' % (anc.get_absolute_url(), anc.title))
-    bc.append(page.title)
-    return "".join(bc)
 
+    bc = [(anc.get_absolute_url(), anc.short_title()) for anc in ancs]
 
+    if include_self:
+        bc.append((None, page.short_title()))
+
+    return {"trail": bc}
+
+# ------------------------------------------------------------------------
 @register.filter
 def is_parent_of(page1, page2):
     """
@@ -179,8 +258,10 @@ def is_parent_of(page1, page2):
 
     return page1.tree_id == page2.tree_id and page1.lft < page2.lft and page1.rght > page2.rght
 
-
+# ------------------------------------------------------------------------
 @register.filter
 def is_equal_or_parent_of(page1, page2):
     return page1.tree_id == page2.tree_id and page1.lft <= page2.lft and page1.rght >= page2.rght
 
+# ------------------------------------------------------------------------
+# ------------------------------------------------------------------------
