@@ -95,7 +95,7 @@ class TranslatedObjectMixin(object):
             language_code = translation.get_language()
 
         key = '-'.join(['%s' % s for s in
-            self.__class__.__name__,
+            self._meta.db_table,
             self.id,
             language_code,
             ])
@@ -106,6 +106,8 @@ class TranslatedObjectMixin(object):
             trans = self._get_translation_object(self.translations.all(), language_code)
             cache.set(key, trans)
 
+        # Assign self to prevent additional database queries
+        trans.parent = self
         return trans
 
     translation = property(get_translation)
@@ -136,7 +138,8 @@ def Translation(model):
 
     class Inner(models.Model):
         parent = models.ForeignKey(model, related_name='translations')
-        language_code = models.CharField(_('language'), max_length=10, choices=settings.LANGUAGES)
+        language_code = models.CharField(_('language'), max_length=10,
+		choices=settings.LANGUAGES, default=settings.LANGUAGES[0][0])
 
         class Meta:
             abstract = True
