@@ -1,22 +1,23 @@
 from django.contrib.auth.decorators import permission_required
 from django.shortcuts import get_object_or_404, render_to_response
 from django.template import RequestContext
+from django.utils.cache import add_never_cache_headers
+
 
 from feincms.module.page.models import Page
 
 
 def _build_page_response(page, request):
-    response = page.setup_request(request)
-    
-    if response is None:
-        extra_context = request._feincms_extra_context
-        response = render_to_response(page.template.path, {
-            'feincms_page': page,
-            }, context_instance=RequestContext(request, extra_context))
-
-    return response
+    extra_context = request._feincms_extra_context
+    return render_to_response(page.template.path, {
+        'feincms_page': page,
+        }, context_instance=RequestContext(request, extra_context))
 
 def build_page_response(page, request):
+    response = page.setup_request(request)
+    if response:
+        return response
+
     response = _build_page_response(page, request)
     page.finalize_response(request, response)
     return response
@@ -32,6 +33,9 @@ def handler(request, path=None):
 
     response = build_page_response(page, request)
 
+    if request.session and request.session.get('frontend_editing', False):
+        add_never_cache_headers(response)
+
     return response
 
 
@@ -43,4 +47,8 @@ def preview_handler(request, page_id):
     is active or expired. To balance that, it requires a logged in user.
     """
     page = get_object_or_404(Page, pk=page_id)
+    # Note: Does not call finalize_response so that response processors
+    # will not kick in, as they might cache the page or do something
+    # equally inappropriate. We are just previewing the page, move along.
+    page.setup_request(request)
     return _build_page_response(page, request)
